/*
 *   Copyright 2017-2020, Schlumberger
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package org.opengroup.osdu.notification.logging;

import org.opengroup.osdu.core.common.http.ResponseHeadersFactory;
import org.opengroup.osdu.core.common.logging.ILogger;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.core.common.model.http.Request;
import org.opengroup.osdu.notification.di.RequestInfoExt;
import static org.opengroup.osdu.core.common.http.ResponseHeaders.STANDARD_RESPONSE_HEADERS;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import javax.servlet.*;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.time.Duration;

@Component
public class ResponseLogFilter implements Filter {

    @Autowired
    private RequestInfoExt requestInfoExt;

    @Autowired
    private ILogger logger;

    private ResponseHeadersFactory responseHeadersFactory = new ResponseHeadersFactory();

    // defaults to * for any front-end, string must be comma-delimited if more than one domain
    @Value("${ACCESS_CONTROL_ALLOW_ORIGIN_DOMAINS:*}")
    String ACCESS_CONTROL_ALLOW_ORIGIN_DOMAINS;

    @Override
    public void init(FilterConfig filterConfig) {
        //do nothing
    }

    @Override
    public void doFilter(ServletRequest servletRequest, ServletResponse servletResponse, FilterChain filterChain) throws IOException, ServletException {
        HttpServletRequest httpServletRequest = (HttpServletRequest) servletRequest;
        HttpServletResponse httpServletResponse = (HttpServletResponse) servletResponse;

        DpsHeaders dpsHeaders = requestInfoExt.getHeaders();
        dpsHeaders.addCorrelationIdIfMissing();
        setResponseHeaders(httpServletResponse);
        httpServletResponse.addHeader(DpsHeaders.CORRELATION_ID, dpsHeaders.getCorrelationId());

        if (httpServletRequest.getMethod().equalsIgnoreCase("OPTIONS")) {
            httpServletResponse.setStatus(200);
        } else {
            filterChain.doFilter(servletRequest, servletResponse);
        }

        long startTime;
        HttpServletRequest request = (HttpServletRequest) servletRequest;
        Object property = request.getServletContext().getAttribute("starttime");
        if (property == null)
            startTime = System.currentTimeMillis();
        else
            startTime = (long) property;

        logger.request("notification.request", Request.builder()
                .requestMethod(request.getMethod())
                .latency(Duration.ofMillis(System.currentTimeMillis() - startTime))
                .requestUrl(request.getRequestURI())
                .Status(((HttpServletResponse) servletResponse).getStatus())
                .ip(requestInfoExt.getUserIp())
                .build(), dpsHeaders.getHeaders());
    }

    @Override
    public void destroy() { }

    private void setResponseHeaders(HttpServletResponse httpServletResponse) {
        Map<String, String> responseHeaders = responseHeadersFactory.getResponseHeaders(ACCESS_CONTROL_ALLOW_ORIGIN_DOMAINS);
        for (Map.Entry<String, String> header : responseHeaders.entrySet()) {
            httpServletResponse.addHeader(header.getKey(), header.getValue());
        }
    }
}
