// Copyright © Amazon Web Services
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package org.opengroup.osdu.notification.provider.aws.queue.impl;

import com.amazonaws.services.dynamodbv2.datamodeling.DynamoDBDeleteExpression;
import com.amazonaws.services.sqs.AmazonSQS;
import com.amazonaws.services.sqs.model.ChangeMessageVisibilityBatchRequestEntry;
import com.amazonaws.services.sqs.model.Message;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperFactory;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperV2;
import org.opengroup.osdu.core.common.http.HttpResponse;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.notification.provider.aws.model.FailedNotificationDoc;
import org.opengroup.osdu.notification.provider.aws.model.NotificationResult;
import org.opengroup.osdu.notification.provider.aws.model.RetryProcessResult;
import org.opengroup.osdu.notification.provider.aws.queue.NotificationRetryQueueService;
import org.opengroup.osdu.notification.provider.aws.repository.SubscriptionRepository;
import org.opengroup.osdu.notification.provider.aws.service.AwsNotificationHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.stream.Collectors;

@Service
public class NotificationRetryQueueServiceImpl implements NotificationRetryQueueService {

    private static final Logger logger = LoggerFactory.getLogger(NotificationRetryQueueServiceImpl.class);

    private static final int SQS_MIN_VISIBILITY_TIMEOUT = 30;

    private static final int SQS_MAX_VISIBILITY_TIMEOUT = 43200;

    private static final String FAILED_NOTIFICATION_RECORD_ID = "failedNotificationRecordId";

    private AwsNotificationHandler notificationHandler;

    private DynamoDBQueryHelperFactory dynamoDBQueryHelperFactory;

    private SubscriptionRepository subscriptionRepository;

    @Value("${aws.dynamodb.failedNotificationTable.ssmPath}")
    private String failedNotificationTablePath;

    @Value("${aws.sqs.visibilityTimeout:300}")
    private Integer visibilityTimeout;

    public NotificationRetryQueueServiceImpl(AwsNotificationHandler notificationHandler, DynamoDBQueryHelperFactory dynamoDBQueryHelperFactory,
                                             SubscriptionRepository subscriptionRepository) {
        this.notificationHandler = notificationHandler;
        this.dynamoDBQueryHelperFactory = dynamoDBQueryHelperFactory;
        this.subscriptionRepository = subscriptionRepository;
    }

    @Override
    public List<RetryProcessResult> processNotificationMessages(List<Message> messages) {
        Map<String, List<Message>> messagesBySubscription = messages.stream().collect(Collectors.groupingBy(
                msg -> String.join(":", msg.getMessageAttributes().get("data-partition-id").getStringValue(), msg.getMessageAttributes().get("subscriptionId").getStringValue()),
                HashMap::new, Collectors.toCollection(ArrayList::new)));

        return messagesBySubscription.entrySet().parallelStream().map(entry -> {
            try {
                Optional<Subscription> subscription = subscriptionRepository.getSubscriptionById(entry.getValue().get(0).getMessageAttributes().get("subscriptionId").getStringValue(),
                        entry.getValue().get(0).getMessageAttributes().get("data-partition-id").getStringValue());
                if (!subscription.isPresent()) {
                    logger.info("Subscription with id {} not found for dataPartitionId {}",
                            entry.getValue().get(0).getMessageAttributes().get("subscriptionId").getStringValue(),
                            entry.getValue().get(0).getMessageAttributes().get("data-partition-id").getStringValue());
                    deleteDbRecords(entry.getValue());
                    return entry.getValue().stream().map(msg -> new RetryProcessResult(msg, NotificationResult.ACK)).collect(Collectors.toList());
                }
                return processMessagesBySubscription(subscription.get(), entry.getValue());
            } catch (Exception e) {
                logger.error("Exception while trying to send retry notifications:", e);
                return entry.getValue().stream().map(msg -> new RetryProcessResult(msg, NotificationResult.NACK)).collect(Collectors.toList());
            }
        }).flatMap(Collection::stream).collect(Collectors.toList());
    }

    @Override
    public List<RetryProcessResult> processMessagesBySubscription(Subscription subscription, List<Message> messages) {
        DynamoDBQueryHelperV2 dynamoDBQueryHelper = dynamoDBQueryHelperFactory.getQueryHelperUsingSSM(failedNotificationTablePath);
        boolean hasPreviousFailed = false;
        List<RetryProcessResult> notificationResults = new ArrayList<>();

        for (Message message : messages) {
            RetryProcessResult processResult = new RetryProcessResult();
            processResult.setMessage(message);

            logger.debug("Processing Notification for messageId: {}, Subscription Endpoint: {}", message.getMessageId(), subscription.getPushEndpoint());
            if (hasPreviousFailed) {
                processResult.setResult(NotificationResult.NACK);
                notificationResults.add(processResult);
                continue;
            }
            Map<String, String> messageAttributes = message.getMessageAttributes().entrySet().stream().collect(
                    Collectors.toMap(Entry::getKey, attribute -> attribute.getValue().getStringValue()));
            try {
                HttpResponse response = notificationHandler.notifySubscriber(subscription, message.getBody(), messageAttributes);
                if (!response.isSuccessCode()) {
                    processResult.setResult(NotificationResult.NACK);
                    hasPreviousFailed = true;
                } else {
                    processResult.setResult(NotificationResult.ACK);
                    FailedNotificationDoc doc = dynamoDBQueryHelper.loadByPrimaryKey(FailedNotificationDoc.class,
                            message.getMessageAttributes().get(FAILED_NOTIFICATION_RECORD_ID).getStringValue());
                    if (doc != null) {
                        FailedNotificationDoc objectToDelete = new FailedNotificationDoc();
                        objectToDelete.setId(message.getMessageAttributes().get(FAILED_NOTIFICATION_RECORD_ID).getStringValue());
                        DynamoDBDeleteExpression deleteExpression = new DynamoDBDeleteExpression()
                                .withConditionExpression("attribute_exists(id)");
                        dynamoDBQueryHelper.deleteByObjectWithCondition(objectToDelete, deleteExpression);
                    }
                }
            } catch (Exception e) {
                logger.error("Failed to notify subscriber : ", e);
                processResult.setResult(NotificationResult.NACK);
                hasPreviousFailed = true;
            }
            notificationResults.add(processResult);
        }
        return notificationResults;
    }

    public void changeMessageVisibilityTimeout(AmazonSQS sqsClient, String sqsQueueUrl, List<Message> messages) {
        List<ChangeMessageVisibilityBatchRequestEntry> entries = new ArrayList<>();
        for (Message message : messages) {
            int timeout = exponentialTimeOutWindow(parseIntOrDefault(message.getAttributes().get("ApproximateReceiveCount"), 0));
            if (timeout < SQS_MIN_VISIBILITY_TIMEOUT || timeout > SQS_MAX_VISIBILITY_TIMEOUT) {
                timeout = timeout < SQS_MIN_VISIBILITY_TIMEOUT ? SQS_MIN_VISIBILITY_TIMEOUT : SQS_MAX_VISIBILITY_TIMEOUT;
            }
            entries.add(
                    new ChangeMessageVisibilityBatchRequestEntry(message.getMessageId(), message.getReceiptHandle())
                            .withVisibilityTimeout(timeout));
            if (entries.size() == 10) {
                sqsClient.changeMessageVisibilityBatch(sqsQueueUrl, entries);
                entries.clear();
            }
        }

        if (!entries.isEmpty()) {
            sqsClient.changeMessageVisibilityBatch(sqsQueueUrl, entries);
        }
    }

    private void deleteDbRecords(List<Message> messages) {
        DynamoDBQueryHelperV2 dynamoDBQueryHelper = dynamoDBQueryHelperFactory.getQueryHelperUsingSSM(failedNotificationTablePath);
        messages.parallelStream().forEach(message -> {
            FailedNotificationDoc doc = dynamoDBQueryHelper.loadByPrimaryKey(FailedNotificationDoc.class,
                    message.getMessageAttributes().get(FAILED_NOTIFICATION_RECORD_ID).getStringValue());
            try {
                FailedNotificationDoc objectToDelete = new FailedNotificationDoc();
                objectToDelete.setId(message.getMessageAttributes().get(FAILED_NOTIFICATION_RECORD_ID).getStringValue());
                DynamoDBDeleteExpression deleteExpression = new DynamoDBDeleteExpression()
                        .withConditionExpression("attribute_exists(id)");
                dynamoDBQueryHelper.deleteByObjectWithCondition(objectToDelete, deleteExpression);
            } catch (Exception e) {
                logger.info("Failed to delete: {}", e.getMessage());
            }
        });
    }

    private Integer exponentialTimeOutWindow(int receiveCount) {
        if (receiveCount <= 10) {
            return visibilityTimeout;
        }
        if (receiveCount <= 20) {
            return visibilityTimeout * 2;
        }
        if (receiveCount <= 30) {
            return visibilityTimeout * 4;
        }
        if (receiveCount <= 40) {
            return visibilityTimeout * 8;
        }
        if (receiveCount <= 50) {
            return visibilityTimeout * 16;
        }
        if (receiveCount <= 60) {
            return visibilityTimeout * 32;
        }
        return 43200;
    }

    private Integer parseIntOrDefault(String toParse, int defaultValue) {
        try {
            return Integer.parseInt(toParse);
        } catch (NumberFormatException e) {
            return defaultValue;
        }
    }

}
