// Copyright © Amazon Web Services
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package org.opengroup.osdu.notification.provider.aws.repository;

import com.amazonaws.services.dynamodbv2.datamodeling.PaginatedQueryList;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperFactory;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperV2;
import org.opengroup.osdu.core.common.model.notification.GsaSecret;
import org.opengroup.osdu.core.common.model.notification.GsaSecretValue;
import org.opengroup.osdu.core.common.model.notification.HmacSecret;
import org.opengroup.osdu.core.common.model.notification.Secret;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.notification.provider.aws.model.SubscriptionDoc;
import org.opengroup.osdu.notification.provider.aws.security.KmsHelper;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Repository;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

@Repository
public class SubscriptionRepository {

    private static final String GSA_SECRET = "GSA";

    private DynamoDBQueryHelperFactory dynamoDBQueryHelperFactory;

    private KmsHelper kmsHelper;

    @Value("${aws.dynamodb.subscriptionTable.ssm.relativePath}")
    String subscriptionTableRelativePath;

    public SubscriptionRepository(DynamoDBQueryHelperFactory dynamoDBQueryHelperFactory, KmsHelper kmsHelper) {
        this.dynamoDBQueryHelperFactory = dynamoDBQueryHelperFactory;
        this.kmsHelper = kmsHelper;
    }

    public List<Subscription> getAllSubscriptionsByDataPartition(String dataPartitionId) {
        DynamoDBQueryHelperV2 queryHelper = dynamoDBQueryHelperFactory.getQueryHelperForPartition(dataPartitionId, subscriptionTableRelativePath);
        SubscriptionDoc gsiQuery = new SubscriptionDoc();
        gsiQuery.setDataPartitionId(dataPartitionId);
        PaginatedQueryList<SubscriptionDoc> results = queryHelper.queryByGSI(SubscriptionDoc.class, gsiQuery);

        if(results == null){
            return Collections.emptyList();
        }
        return results.stream().map(subsDoc -> {
            String secretValue = kmsHelper.decrypt(subsDoc.getSecretValue(), dataPartitionId);
            Secret secret = createSecret(subsDoc.getSecretType(), secretValue);
            return SubscriptionDoc.mapTo(subsDoc, secret);
        }).collect(Collectors.toList());
    }

    public Optional<Subscription> getSubscriptionById(String subscriptionId, String dataPartitionId) {
        DynamoDBQueryHelperV2 queryHelper = dynamoDBQueryHelperFactory.getQueryHelperForPartition(dataPartitionId, subscriptionTableRelativePath);
        SubscriptionDoc subsDoc = queryHelper.loadByPrimaryKey(SubscriptionDoc.class, subscriptionId);
        if (subsDoc == null) {
            return Optional.empty();
        }
        String secretValue = kmsHelper.decrypt(subsDoc.getSecretValue(), dataPartitionId);
        Secret secret = createSecret(subsDoc.getSecretType(), secretValue);
        return Optional.of(SubscriptionDoc.mapTo(subsDoc, secret));
    }

    private Secret createSecret(String secretType, String secretValue) {
        Secret secret;
        if (secretType.equals(GSA_SECRET)) {
            GsaSecret gsaSecret = new GsaSecret();
            String[] splitSecret = secretValue.split("`");
            gsaSecret.setValue(new GsaSecretValue(splitSecret[0], splitSecret[1]));
            secret = gsaSecret;
        } else {
            HmacSecret hmacSecret = new HmacSecret();
            hmacSecret.setValue(secretValue);
            secret = hmacSecret;
        }
        return secret;
    }

}