// Copyright © 2020 Amazon Web Services
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package org.opengroup.osdu.notification.provider.aws.security;


import com.amazonaws.services.kms.AWSKMS;
import com.amazonaws.services.kms.model.DecryptRequest;
import com.amazonaws.services.kms.model.EncryptRequest;
import com.amazonaws.services.simplesystemsmanagement.model.InternalServerErrorException;
import lombok.AccessLevel;
import lombok.Setter;
import org.apache.http.HttpStatus;
import org.opengroup.osdu.core.aws.ssm.K8sLocalParameterProvider;
import org.opengroup.osdu.core.aws.ssm.K8sParameterNotFoundException;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import java.nio.ByteBuffer;
import java.util.Collections;

@Component
public class KmsHelper {

    @Value("${aws.region}")
    @Setter(AccessLevel.PROTECTED)
    public String amazonRegion;

    @Autowired
    private DpsHeaders dpsHeaders;

    @Autowired
    private JaxRsDpsLog logger;

    @Value("${aws.kms.endpoint}")
    public String kmsEndpoint;

    private AWSKMS kmsClient;

    private String kmsKeyId;

    @PostConstruct
    public void init() {
        KmsConfig kmsConfig = new KmsConfig(kmsEndpoint, amazonRegion);
        kmsClient = kmsConfig.awsKMS();
        try {
            K8sLocalParameterProvider provider = new K8sLocalParameterProvider();
            kmsKeyId = provider.getParameterAsString("KMS_KEY_ID");
        } catch (K8sParameterNotFoundException e) {
            throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "SSM ParameterNotFoundException", e.getMessage());
        } catch (InternalServerErrorException e) {
            throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "SSM InternalServerErrorException", e.getErrorMessage());

        }

    }



   public ByteBuffer encrypt(String plainTextString) {


        EncryptRequest encReq = new EncryptRequest();
        encReq.setKeyId(kmsKeyId);
        encReq.setPlaintext(ByteBuffer.wrap(plainTextString.getBytes()));
        encReq.setEncryptionContext(Collections.singletonMap("dataPartitionId", dpsHeaders.getPartitionId()));
        ByteBuffer ciphertext = kmsClient.encrypt(encReq).getCiphertextBlob();
        return ciphertext;



    }
    public String decrypt(ByteBuffer ciphertext, String dataPartitionId) {

        DecryptRequest decReq = new DecryptRequest();

        decReq.setCiphertextBlob(ciphertext);
        decReq.setEncryptionContext(Collections.singletonMap("dataPartitionId", dataPartitionId));
        ByteBuffer decrypted = kmsClient.decrypt(decReq).getPlaintext();

        String decryptedStr = new String(decrypted.array());
        return decryptedStr;
    }

}
