// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.notification.provider.azure.messageBus;

import com.google.api.client.util.Strings;
import com.microsoft.applicationinsights.TelemetryClient;
import com.microsoft.applicationinsights.telemetry.RequestTelemetry;
import com.microsoft.azure.servicebus.IMessage;
import org.apache.commons.lang3.time.StopWatch;
import org.opengroup.osdu.core.common.http.HttpResponse;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.notification.provider.azure.messageBus.thread.ThreadScopeContextHolder;
import org.opengroup.osdu.notification.provider.azure.models.NotificationContent;
import org.opengroup.osdu.notification.provider.azure.messageBus.extractor.RequestBodyAdapter;
import org.opengroup.osdu.notification.provider.azure.messageBus.thread.ThreadDpsHeaders;
import org.opengroup.osdu.notification.provider.azure.util.MDCContextMap;
import org.opengroup.osdu.notification.service.NotificationHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.MDC;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnExpression;
import org.springframework.stereotype.Component;
import java.util.concurrent.ConcurrentMap;
import java.util.Date;

@Component
@ConditionalOnExpression("${azure.serviceBus.enabled:true} || ${azure.eventGridToServiceBus.enabled:true}")
public class ProcessNotification {
    private final String NOT_ACKNOWLEDGE = "message not acknowledged by client";
    private final static Logger LOGGER = LoggerFactory.getLogger(ProcessNotification.class);
    @Autowired
    private NotificationHandler notificationHandler;
    @Autowired
    private RequestBodyAdapter requestBodyAdapter;
    @Autowired
    private ThreadDpsHeaders dpsHeaders;
    @Autowired
    private MDCContextMap mdcContextMap;

    public void performNotification(IMessage message, String subscriptionName) throws Exception {
        TelemetryClient telemetryClient = new TelemetryClient();
        StopWatch stopWatch = new StopWatch();
        stopWatch.start();        

        try {
            NotificationContent notificationContent = requestBodyAdapter.extractNotificationContent(message, subscriptionName);

            String dataPartitionId = notificationContent.getExtractAttributes().get(DpsHeaders.DATA_PARTITION_ID);
            String correlationId = notificationContent.getExtractAttributes().get(DpsHeaders.CORRELATION_ID);
            String collaborationId = null;
            if (notificationContent.getExtractAttributes().containsKey("x-collaboration"))
                collaborationId = notificationContent.getExtractAttributes().get("x-collaboration");
            ConcurrentMap properties = telemetryClient.getContext().getProperties();
            properties.put("correlation-id",correlationId);
            properties.put("data-partition-id", dataPartitionId);
            if (!Strings.isNullOrEmpty(collaborationId))
                properties.put("x-collaboration", collaborationId);
            MDC.setContextMap(mdcContextMap.getContextMap(correlationId, dataPartitionId, collaborationId));
            dpsHeaders.setThreadContext(dataPartitionId, correlationId, collaborationId);

            HttpResponse response = notificationHandler.notifySubscriber(notificationContent.getNotificationId(),
                    notificationContent.getData(), notificationContent.getExtractAttributes());

            RequestTelemetry requestTelemetry = new RequestTelemetry(
                "SBQueueRequest",
                new Date(),
                stopWatch.getTime(),
                "424",
                false);
            requestTelemetry.setId(message.getMessageId());                    
                                           
            if (!response.isSuccessCode()) {
                telemetryClient.trackRequest(requestTelemetry); 
                throw new Exception(NOT_ACKNOWLEDGE);
            }else{
                requestTelemetry.setResponseCode("200");
                requestTelemetry.setSuccess(true);
                telemetryClient.trackRequest(requestTelemetry);
            }
        } catch (Exception e) {
            throw e;
        } finally {
            ThreadScopeContextHolder.getContext().clear();
            MDC.clear();
            stopWatch.stop();
        }
    }
}
