// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.notification.provider.azure.pubsub;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.base.Preconditions;
import com.google.common.base.Strings;
import com.google.gson.Gson;
import com.google.gson.JsonObject;
import lombok.SneakyThrows;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.notification.provider.azure.models.NotificationContent;
import org.opengroup.osdu.notification.provider.azure.models.HandshakeRequestData;
import org.opengroup.osdu.notification.provider.azure.models.NotificationRecordsChangedData;
import org.opengroup.osdu.notification.provider.azure.models.NotificationEventGridRequest;
import org.opengroup.osdu.notification.provider.interfaces.IPubsubRequestBodyExtractor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Component;
import org.springframework.web.context.annotation.RequestScope;

import javax.servlet.http.HttpServletRequest;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;

// TODO : Make error messages as constants.
// TODO : Add TracePoints to exception logs so that monitoring is done well.

@Component
@RequestScope
public class EventGridRequestBodyExtractor implements IPubsubRequestBodyExtractor {

    private static final String SUBSCRIPTION_ID = "Aeg-Subscription-Name";
    private static final String EVENTGRID_VALIDATION_EVENT = "Microsoft.EventGrid.SubscriptionValidationEvent";
    private static final Gson GSON = new Gson();
    private static final ObjectMapper objectMapper = new ObjectMapper();

    private final JsonObject root = null;
    private final HttpServletRequest httpServletRequest;
    private final JaxRsDpsLog logger;

    private final NotificationEventGridRequest notificationRequest;
    private NotificationRecordsChangedData notificationRecordsChangedData;
    private HandshakeRequestData handshakeRequestData;
    private boolean isHandshakeRequest;
    private NotificationContent notificationContent;

    @Autowired
    public EventGridRequestBodyExtractor(HttpServletRequest httpServletRequest, JaxRsDpsLog log) {
        this.httpServletRequest = httpServletRequest;
        this.logger = log;
        this.notificationRequest = extractNotificationRequestFromHttpRequest();
    }

    /**
     * Extracts the attributes from the request that are filled in by publisher of the message.
     *
     * @return Request Attributes Map
     * @throws AppException
     */
    public Map<String, String> extractAttributesFromRequestBody() {
        if (isHandshakeRequest) {
            logger.error("Invalid Event Grid Message. Is a handshake request");
            return null;
        }
        Map<String, String> attributes = new HashMap<>();
        attributes.put("correlation-id", this.notificationRecordsChangedData.getCorrelationId());
        attributes.put("data-partition-id", this.notificationRecordsChangedData.getDataPartitionId());
        attributes.put("account-id", this.notificationRecordsChangedData.getAccountId());
        return attributes;
    }

    /**
     * Extracts the data from the request that are filled in by publisher of the message,
     *
     * @return Request Data String
     * @throws AppException
     */
    public String extractDataFromRequestBody() {
        if (isHandshakeRequest) {
            logger.error("Invalid Event Grid Message. Is a handshake request");
            return null;
        }
        return notificationRecordsChangedData.getData().toString();
    }

    /**
     * Extracts the notificationId from the request that are filled in by EventGrid.
     *
     * @return Request NotificationId String.
     * @throws AppException
     */
    public String extractNotificationIdFromRequestBody() {
        String subscriptionId = httpServletRequest.getHeader(SUBSCRIPTION_ID);
        if (Strings.isNullOrEmpty(subscriptionId)) {
            logger.error("Invalid Event Grid Message. Subscription Id is null or empty");
            throw new AppException(HttpStatus.BAD_REQUEST.value(), "Invalid Event Grid Message", "Subscription ID not found");
        }

        return subscriptionId.toLowerCase();
    }

    /**
     * Checks if the request is for handshake.
     *
     * @return Request Type Boolean
     * @throws AppException
     */
    public boolean isHandshakeRequest() {
        return this.isHandshakeRequest;
    }

    /**
     * Return ValidationCode
     *
     * @return Request Type Boolean
     * @throws AppException
     */
    public String getValidationCodeForHandshake() {
        if (!isHandshakeRequest) {
            logger.error("Invalid Event Grid Message. Is not a handshake request");
            return null;
        }
        return this.handshakeRequestData.getValidationCode();
    }

    /**
     * Utility method that extracts the core content in the request.
     * This is what the publisher will publish.
     * <p>
     * The attributes are validated.
     *
     * @return NotificationRequest Object
     * @throws AppException
     */
    @SneakyThrows
    // TODO : @komakkar sanitize the exceptions to match the SpringExceptionMapper and throw ValidationException
    private NotificationEventGridRequest extractNotificationRequestFromHttpRequest() {
        NotificationEventGridRequest notificationRequest = null;
        if (this.notificationRequest == null && this.httpServletRequest.getMethod().equalsIgnoreCase("post")) {
            try {
                String requestBody = getBody(this.httpServletRequest);
                NotificationEventGridRequest[] notificationRequestArray = GSON.fromJson(requestBody, NotificationEventGridRequest[].class);
                notificationRequest = notificationRequestArray[0];

                this.isHandshakeRequest = notificationRequest.getEventType().equals(EVENTGRID_VALIDATION_EVENT);
                if (isHandshakeRequest) {
                    extractHandshakeData(notificationRequest);
                } else {
                    extractNotificationData(notificationRequest);
                }
            } catch (Exception e) {
                throw new AppException(HttpStatus.BAD_REQUEST.value(), "Request payload parsing error",
                        "Unable to parse request payload.", "Request contents are null or empty", e);
            }
        }
        return notificationRequest;
    }

    private void extractHandshakeData(NotificationEventGridRequest notificationRequest) {
        this.handshakeRequestData = GSON.fromJson(notificationRequest.getData(), HandshakeRequestData.class);
        Preconditions.checkNotNull(this.handshakeRequestData.getValidationCode(), "Request payload parsing error handshkae");
    }

    private void extractNotificationData(NotificationEventGridRequest notificationRequest) {
        NotificationRecordsChangedData notificationRecordsChangedData = GSON.fromJson(notificationRequest.getData(), NotificationRecordsChangedData.class);
        verifyNotificationData(notificationRecordsChangedData);
        this.notificationRecordsChangedData = notificationRecordsChangedData;
    }

    // TODO: Clean up for using @NonNull Lombok.
    // The Gson deserialisation happens through constructor made out of reflection,
    // The unsafe initialisation inspires us to use Jackson
    //
    // Making Gson use a custom constructor, involves boilerplating, hence adding precondition checks. https://github.com/google/gson/blob/master/UserGuide.md#TOC-Writing-a-Deserializer
    // The required condition is not operation as of now.https://github.com/google/gson/issues/61
    private void verifyNotificationData(NotificationRecordsChangedData notificationRecordsChangedData) {
        Preconditions.checkNotNull(notificationRecordsChangedData, "Request payload parsing error");
        Preconditions.checkNotNull(notificationRecordsChangedData.getData(), "Request payload parsing error");
        Preconditions.checkNotNull(notificationRecordsChangedData.getCorrelationId(), "Request payload parsing error");
        Preconditions.checkNotNull(notificationRecordsChangedData.getDataPartitionId(), "Request payload parsing error");
    }

    private String getBody(HttpServletRequest request) throws IOException {
        String body;
        StringBuilder stringBuilder = new StringBuilder();
        BufferedReader bufferedReader = null;
        try {
            InputStream inputStream = request.getInputStream();
            if (inputStream != null) {
                bufferedReader = new BufferedReader(new InputStreamReader(inputStream));
                char[] charBuffer = new char[128];
                int bytesRead = -1;
                while ((bytesRead = bufferedReader.read(charBuffer)) > 0) {
                    stringBuilder.append(charBuffer, 0, bytesRead);
                }
            } else {
                stringBuilder.append("");
            }
        } finally {
            if (bufferedReader != null) {
                bufferedReader.close();
            }
        }
        body = stringBuilder.toString();
        return body;
    }
}
