/*
 *  Copyright 2020-2023 Google LLC
 *  Copyright 2020-2023 EPAM Systems, Inc
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.opengroup.osdu.notification.provider.gcp.pubsub;

import java.util.List;
import java.util.stream.Collectors;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.http.HttpStatus;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.core.common.notification.ISubscriptionFactory;
import org.opengroup.osdu.core.common.notification.ISubscriptionService;
import org.opengroup.osdu.core.common.notification.SubscriptionException;
import org.opengroup.osdu.core.gcp.oqm.model.OqmTopic;
import org.opengroup.osdu.notification.provider.gcp.service.DpsHeadersProvider;
import org.opengroup.osdu.notification.provider.gcp.service.SubscriptionServiceGc;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Component;

/**
 * GC implementation of {@link org.opengroup.osdu.notification.service.SubscriptionHandler}.
 * The current implementation differs in non-request spring scope and handling OQM message instead of HTTP requests.
 * Also, this implementation supports additional topics and subscription requests to Register service.
 */
@Component
@RequiredArgsConstructor
@ConditionalOnProperty(name = "oqmDriver")
@Slf4j
public class OqmSubscriptionHandler {

    private final ISubscriptionFactory registerClientFactory;
    private final DpsHeadersProvider dpsHeadersProvider;
    private final SubscriptionServiceGc subscriptionServiceGc;

    public List<OqmTopic> getTopics(String dataPartitionId) {
        DpsHeaders headers = dpsHeadersProvider.getDpsHeaders(dataPartitionId);
        ISubscriptionService registerClient = registerClientFactory.create(headers);

        try {
            return registerClient.getTopics()
                    .stream()
                    .map(topic -> OqmTopic.builder()
                            .name(topic.getName())
                            .build())
                    .collect(Collectors.toList());
        } catch (SubscriptionException se) {
            log.error("Failed to get topics from register for partition id: {}. Response: {}", dataPartitionId, se.getResponse().getBody());
            throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "Failed to get topics from register", se.getResponse().toString(), se);
        }
    }

    public List<Subscription> getAllSubscriptionInfos(String dataPartitionId) {
        try {
            return subscriptionServiceGc.getAll(dataPartitionId);
        } catch (SubscriptionException e) {
            throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "Failed to get all subscription from register",
                    "Received subscription exception: ", e);
        }
    }

    public List<Subscription> getSubscriptionsById(String dataPartitionId, String subscriptionId) {
        try {
            DpsHeaders headers = dpsHeadersProvider.getDpsHeaders(dataPartitionId);
            ISubscriptionService registerClient = registerClientFactory.create(headers);
            return registerClient.query(subscriptionId);
        } catch (SubscriptionException se) {
            throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "Server error", "Unexpected error while sending request", se);
        }
    }
}