/*
 *  Copyright 2020-2023 Google LLC
 *  Copyright 2020-2023 EPAM Systems, Inc
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.opengroup.osdu.notification.provider.gcp.pubsub.receiver;

import java.util.List;
import java.util.Map;
import java.util.Optional;
import lombok.extern.slf4j.Slf4j;
import org.apache.http.HttpStatus;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.core.gcp.oqm.model.OqmAckReplier;
import org.opengroup.osdu.core.gcp.oqm.model.OqmMessage;
import org.opengroup.osdu.core.gcp.oqm.model.OqmMessageReceiver;
import org.opengroup.osdu.notification.provider.gcp.config.ExternalSubscriptionsManager;
import org.opengroup.osdu.notification.provider.gcp.model.ExternalSubscriptions;
import org.opengroup.osdu.notification.provider.gcp.thread.ThreadScopeContextHolder;

@Slf4j
public class OqmControlTopicReceiver implements OqmMessageReceiver {

    private static final String SUBSCRIPTION_CREATED = "Subscription Created";
    private static final String SUBSCRIPTION_UPDATED = "Subscription Updated";
    private static final String SUBSCRIPTION_DELETED = "Subscription Deleted";

    private final ExternalSubscriptionsManager externalSubscriptionsManager;

    public OqmControlTopicReceiver(ExternalSubscriptionsManager externalSubscriptionsManager) {
        this.externalSubscriptionsManager = externalSubscriptionsManager;
    }

    @Override
    public void receiveMessage(OqmMessage message, OqmAckReplier replier) {
        try {
            handleMessage(message);
        } catch (Exception e) {
            log.error("OQM | Control topic | Message not acknowledged by client", e);
            replier.nack();
        } finally {
            ThreadScopeContextHolder.currentThreadScopeAttributes().clear();
        }

        log.debug("OQM | Control topic | Message acknowledged by client");
        replier.ack();
    }

    private void handleMessage(OqmMessage oqmMessage) {
        String pubsubMessage = oqmMessage.getData();
        Map<String, String> headerAttributes = oqmMessage.getAttributes();
        String subscriptionId = headerAttributes.get("subscription-id");
        String serviceTopic = headerAttributes.get("topic");
        String dataPartitionId = headerAttributes.get("data-partition-id");
        log.debug("OQM | Control topic | Received message: `{}` for service topic: `{}` with Sub ID: `{}`",
                pubsubMessage, serviceTopic, subscriptionId);

        List<Subscription> subscriptionInfos = Optional.ofNullable(
                externalSubscriptionsManager.getExternalSubscriptions(dataPartitionId))
                .orElseThrow(() -> new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "Server error", "OQM | Subscription info cache was not initialized"))
                .getSubscriptions();
        Optional<Subscription> cachedInfo = subscriptionInfos.stream()
                .filter(info -> subscriptionId.equals(info.getNotificationId()) && serviceTopic.equals(info.getTopic()))
                .findFirst();

        switch (pubsubMessage) {
            case SUBSCRIPTION_CREATED:
            case SUBSCRIPTION_UPDATED:
                cachedInfo.ifPresent(subscriptionInfos::remove);
                Subscription freshSubscriptionInfo = externalSubscriptionsManager.getSubscription(dataPartitionId, subscriptionId, serviceTopic);
                if (freshSubscriptionInfo != null) {
                    subscriptionInfos.add(freshSubscriptionInfo);
                    externalSubscriptionsManager.updateExternalSubscriptionsCache(dataPartitionId, ExternalSubscriptions.builder().subscriptions(subscriptionInfos).build());
                    log.debug("OQM | Control topic | Subscription info with Sub ID: `{}` and topic: `{}` was added to cache.", subscriptionId, serviceTopic);
                }
                break;
            case SUBSCRIPTION_DELETED:
                if (cachedInfo.isPresent()) {
                    subscriptionInfos.remove(cachedInfo.get());
                    externalSubscriptionsManager.updateExternalSubscriptionsCache(dataPartitionId, ExternalSubscriptions.builder().subscriptions(subscriptionInfos).build());
                    log.debug("OQM | Control topic | Subscription info with Sub ID: `{}` and topic: `{}` was removed from cache.", subscriptionId, serviceTopic);
                } else {
                    log.warn("OQM | Control topic requested to delete subscription with id: `{}`, but is was not found in cache.", subscriptionId);
                }
                break;
            default:
                throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "Not supported", "OQM | Control topic | Received unsupported control topic operation: " + pubsubMessage);
        }
    }
}