/*
 *   Copyright 2017-2020, Schlumberger
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package org.opengroup.osdu.notification.secutity;

import com.google.api.client.googleapis.auth.oauth2.GoogleIdToken;
import com.google.api.client.googleapis.auth.oauth2.GoogleIdTokenVerifier;
import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.notification.provider.gcp.security.GoogleServiceAccountValidatorGenerator;
import org.opengroup.osdu.notification.provider.gcp.security.GoogleServiceAccountValidatorImpl;
import org.powermock.modules.junit4.PowerMockRunner;

import java.io.IOException;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

@RunWith(PowerMockRunner.class)
public class GoogleServiceAccountValidatorImplTests {
    private static final String TEST_JWT = "testjwt";
    private static final String TEST_USER_IDENTITY = "testidentity";

    @Mock
    private JaxRsDpsLog log;
    @Mock
    private GoogleServiceAccountValidatorGenerator verifierGenerator;
    @Mock
    private GoogleIdTokenVerifier verifier;
    @Mock
    private GoogleIdToken idToken;
    @Mock
    private GoogleIdToken.Payload payload;
    @InjectMocks
    private GoogleServiceAccountValidatorImpl sut;

    @Test
    public void should_returnTrue_when_tokenValidAndUserIdentityCorrect() throws Exception{
        when(this.verifierGenerator.getVerifier(any(), any())).thenReturn(this.verifier);
        when(this.verifier.verify(TEST_JWT)).thenReturn(this.idToken);
        when(this.idToken.getPayload()).thenReturn(this.payload);
        when(this.payload.getEmail()).thenReturn(TEST_USER_IDENTITY);
        when(this.payload.getEmailVerified()).thenReturn(Boolean.TRUE);
        Assert.assertTrue(this.sut.isValidServiceAccount(TEST_JWT, TEST_USER_IDENTITY));
    }

    @Test
    public void should_returnFalse_when_tokenInvalid() throws Exception {
        when(this.verifierGenerator.getVerifier(any(), any())).thenReturn(this.verifier);
        when(this.verifier.verify(TEST_JWT)).thenReturn(null);
        Assert.assertFalse(this.sut.isValidServiceAccount(TEST_JWT, TEST_USER_IDENTITY));
    }

    @Test
    public void should_returnFalse_when_tokenValidAndUserIdentityIncorrect() throws Exception{
        when(this.verifierGenerator.getVerifier(any(), any())).thenReturn(this.verifier);
        when(this.verifier.verify(TEST_JWT)).thenReturn(this.idToken);
        when(this.idToken.getPayload()).thenReturn(this.payload);
        when(this.payload.getEmail()).thenReturn("wrongIdentity");
        when(this.payload.getEmailVerified()).thenReturn(Boolean.TRUE);
        Assert.assertFalse(this.sut.isValidServiceAccount(TEST_JWT, TEST_USER_IDENTITY));
    }

    @Test
    public void should_returnFalse_when_tokenValidAndUserIdentityCorrect_butEmailNotVerified() throws Exception{
        when(this.verifierGenerator.getVerifier(any(), any())).thenReturn(this.verifier);
        when(this.verifier.verify(TEST_JWT)).thenReturn(this.idToken);
        when(this.idToken.getPayload()).thenReturn(this.payload);
        when(this.payload.getEmail()).thenReturn(TEST_USER_IDENTITY);
        when(this.payload.getEmailVerified()).thenReturn(Boolean.FALSE);
        Assert.assertFalse(this.sut.isValidServiceAccount(TEST_JWT, TEST_USER_IDENTITY));
    }

    @Test
    public void should_logExceptionAndReturnFalse_when_tokenValidationThrowsException() throws Exception{
        when(this.verifierGenerator.getVerifier(any(), any())).thenReturn(this.verifier);
        IOException e = new IOException("invalid token");
        when(this.verifier.verify(TEST_JWT)).thenThrow(e);
        Assert.assertFalse(this.sut.isValidServiceAccount(TEST_JWT, TEST_USER_IDENTITY));
        verify(this.log, times(1)).error("Error when validating google id token", e);
    }
}
