// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.notification.api;

import com.sun.jersey.api.client.ClientResponse;
import org.apache.catalina.connector.Response;
import org.apache.commons.lang3.time.StopWatch;
import org.asynchttpclient.util.Assertions;
import org.junit.*;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.core.common.model.notification.HmacSecret;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.core.common.model.notification.SubscriptionInfo;
import org.opengroup.osdu.core.common.notification.ISubscriptionService;
import org.opengroup.osdu.core.common.notification.SubscriptionAPIConfig;
import org.opengroup.osdu.core.common.notification.SubscriptionException;
import org.opengroup.osdu.core.common.notification.SubscriptionFactory;
import org.opengroup.osdu.notification.util.*;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

public class TestStorageIntegration extends BaseTestTemplate {

    Logger logger= Logger.getLogger("");
    private String subscriptionId = null;
    private String notificationId = null;
    private final String ackSubscriptionId = "AckSubscription";
    private final String ackNotificationId = "testingAcknowledgement";
    private ISubscriptionService subscriptionService;
    private static SubscriptionFactory factory;
    protected static final String LEGAL_TAG = TestUtils.getOsduTenant() + "-test-tag";

    @BeforeClass
    public static void classSetup() throws Exception {
        descriptor = new StorageIntegrationDescriptor();
        //Configure Register Service Client Library
        SubscriptionAPIConfig config = SubscriptionAPIConfig.builder().rootUrl(Config.Instance().RegisterServicePath).build();
        factory = new SubscriptionFactory(config);
    }

    @AfterClass
    public static void classTearDown() throws Exception {
    }

    @Before
    @Override
    public void setup() throws Exception {
        this.testUtils = new AzureTestUtils();
    }

    @Override
    protected void deleteResource() throws Exception {
        subscriptionService.delete(subscriptionId);
    }

    @After
    @Override
    public void tearDown() throws Exception {
        this.testUtils = null;
    }

    @Override
    protected void createResource() throws Exception {
        createResourceInPartition(TestUtils.getOsduTenant());
    }

    private void createResourceInPartition(String partitionId) throws Exception {
        Map<String, String> headers = new HashMap<>();
        headers.put(DpsHeaders.DATA_PARTITION_ID, partitionId);
        headers.put(DpsHeaders.AUTHORIZATION, testUtils.getOpsToken());
        DpsHeaders dpsHeaders = DpsHeaders.createFromMap(headers);
        subscriptionService = factory.create(dpsHeaders);
        logger.info("dps headers created.");

        // Create a new subscription
        Subscription subscription = new Subscription();
        subscription.setName("storage-integration-test-hmac");
        subscription.setDescription("Subscription created for Storage Integration Tests");
        subscription.setTopic(Config.Instance().Topic);
        subscription.setPushEndpoint(Config.Instance().HMACPushUrl);
        HmacSecret secret = new HmacSecret();
        secret.setValue(Config.Instance().hmacSecretValue);
        subscription.setSecret(secret);
        try {
            Subscription subscriptionCreated = subscriptionService.create(subscription);
            notificationId = subscriptionCreated.getNotificationId();
            subscriptionId = subscriptionCreated.getId();
            Config.Instance().NotificationId = notificationId;
        } catch (SubscriptionException e) {
            System.out.println("Subscription exception inner response : " + e.getHttpResponse());
            throw e;
        }
        logger.info("New subscription created");
    }

    private void deleteAckSubscription() throws SubscriptionException {
        try {
            subscriptionService.delete(ackSubscriptionId);
        } catch (SubscriptionException e) {
            if (e.getHttpResponse().getResponseCode() == Response.SC_NOT_FOUND) {
                System.out.println("Test Ack Subscription Not Found for deletion.");
                return;
            }
            Assert.fail("Unable to delete Test Ack Subscription. Deletion Failed." + e);
            throw e;
        }
    }

    @Override
    protected String getArg() {
        return null;
    }

    @Override
    protected String getInvalidArg() {
        return null;
    }

    @Override
    protected int expectedOkResponseCode() {
        return 200;
    }

    @Test
    @Override
    public void should_return20XResponseCode_when_makingValidHttpsRequest() throws Exception {
        try {
            // Creates an actual subscription with a test endpoint i.e REGISTER_CUSTOM_PUSH_URL_HMAC
            this.createResource();
            logger.info("Actual subscription with a test endpoint created successfully");

            // Delay to get the above subscription registered by Notification Service
            TimeUnit.SECONDS.sleep(600);
            logger.info("Subscription registered by Notification service successfully");

            // Delete the Test Ack subscription if there exists any as a part of cleanup.
            // The Test Ack Subscription gets created if the the test endpoint of actual subscription
            // created above receives notification from Notification Service.
            this.deleteAckSubscription();
            //  Delete legal tag
            logger.info("Cleaning up previous Legal Tag if any. ");
            LegalTagUtils.delete(LEGAL_TAG, this.testUtils.getAdminToken());


            //  Create legal tag used in storage record
            logger.info("Creating New legal tag");
            LegalTagUtils.create(LEGAL_TAG, this.testUtils.getAdminToken());


            // Insert a storage Record.This will send a notification to  the event grid/service bus topic
            // subscriptions.From there it goes to the Notification Service.Finally from there it goes to the
            // test endpoint i.e REGISTER_CUSTOM_PUSH_URL_HMAC which creates the Test ack Subscription.
            String URL = Config.Instance().StorageServicePath;
            ClientResponse response = descriptor.run(URL, "", this.testUtils.getAdminToken());
            assertEquals(error(response.getEntity(String.class)), 201, response.getStatus());
            TimeUnit.SECONDS.sleep(120);

            // Verification of Test Ack Subscription creation from the above process.
            StopWatch stopWatch = new StopWatch();
            stopWatch.start();
            // Retrieval of test ack subscription from Cosmos-DB.
            long retryCount = Long.parseLong(Config.Instance().RetryCount);
            long timeOut = Long.parseLong(Config.Instance().TimeOutSeconds);
            SubscriptionInfo ackSubscription = subscriptionService.get(ackSubscriptionId);
            while (ackSubscription == null && retryCount > 0) {
                TimeUnit.SECONDS.sleep(timeOut);
                ackSubscription = subscriptionService.get(ackSubscriptionId);
                logger.info("Retry time: "+retryCount+" to get Test Ack subscription.");
                retryCount--;
            }
            Assertions.assertNotNull(ackSubscription, "Unable to retrieve test ack subscription. Elapsed time in minutes : " + (stopWatch.getTime(TimeUnit.MINUTES)));
            if (ackSubscription != null) {
                assertEquals("Unexpected Test Ack Subscription.", ackNotificationId, ackSubscription.getNotificationId());
            }
            logger.info("IT run successful");
            stopWatch.stop();
        } catch (Exception e) {
            fail("An exception occurred :" + e);
        } finally {
            logger.info("Executing finally block");
            // Delete Actual Subscription
             this.deleteResource();
            // Deletion of Test Ack Subscription
            this.deleteAckSubscription();
            //  Delete legal tag
            logger.info("Deleting legal tag");
            LegalTagUtils.delete(LEGAL_TAG, this.testUtils.getAdminToken());

        }
    }

    /* Keeping the No-op test cases implementation as ignored as the Storage-Notification Integration
    Test scenario does not involves them. */
    @Test
    @Override
    @Ignore
    public void should_return400_when_makingHttpRequestWithoutToken() throws Exception {
    }

    @Test
    @Override
    @Ignore
    public void should_return307_when_makingHttpRequest() throws Exception {
    }

    @Test
    @Override
    @Ignore
    public void should_return401_when_accessingWithNoAccessCredentials() throws Exception {
    }

    @Test
    @Override
    @Ignore
    public void should_return401_when_noAccessOnCustomerTenant() throws Exception {
    }

    @Test
    @Override
    @Ignore
    public void should_return401_when_accessingWithEditorCredentials() throws Exception {
    }

    @Test
    @Override
    @Ignore
    public void should_return401_when_accessingWithAdminCredentials() throws Exception {
    }

    @Test
    @Override
    @Ignore
    public void should_return20X_when_usingCredentialsWithOpsPermission() throws Exception {
    }

    @Test
    @Override
    @Ignore
    public void should_returnOk_when_makingHttpOptionsRequest() throws Exception {
    }
}